@@ -3,7 +3,7 @@
 
   common.js - this file is part of Gerbera.
 
-  Copyright (C) 2018-2021 Gerbera Contributors
+  Copyright (C) 2018-2022 Gerbera Contributors
 
   Gerbera is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 2
@@ -48,6 +48,8 @@ function getPlaylistType(mimetype) {
         return 'm3u';
     if (mimetype == 'audio/x-scpls')
         return 'pls';
+    if (mimetype == 'video/x-ms-asf' || mimetype === 'video/x-ms-asx')
+        return 'asx';
     return '';
 }
 
@@ -356,7 +358,7 @@ function addAudio(obj) {
     // container class is set to UPNP_CLASS_CONTAINER_MUSIC_ALBUM.
 
     const chain = {
-        audio: { title: 'Audio', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
+        audio: { title: 'Audio', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER, metaData: [] },
         allAudio: { title: 'All Audio', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
         allArtists: { title: 'Artists', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
         allGenres: { title: 'Genres', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
@@ -372,15 +374,18 @@ function addAudio(obj) {
         composer: { title: composer, objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER_MUSIC_COMPOSER, metaData: [], res: obj.res, aux: obj.aux, refID: obj.id },
     };
 
+    chain.audio.metaData[M_CONTENT_CLASS] = [ UPNP_CLASS_AUDIO_ITEM ];
     chain.album.metaData[M_ARTIST] = artist;
     chain.album.metaData[M_ALBUMARTIST] = artist;
     chain.album.metaData[M_GENRE] = [ genre ];
     chain.album.metaData[M_DATE] = obj.metaData[M_DATE];
+    chain.album.metaData[M_UPNP_DATE] = obj.metaData[M_UPNP_DATE];
     chain.album.metaData[M_ALBUM] = [ album ];
     chain.artist.metaData[M_ARTIST] = artist;
     chain.artist.metaData[M_ALBUMARTIST] = artist;
     chain.genre.metaData[M_GENRE] = [ genre ];
     chain.year.metaData[M_DATE] = [ date ];
+    chain.year.metaData[M_UPNP_DATE] = [ date ];
     chain.composer.metaData[M_COMPOSER] = [ composer ];
 
     var container = addContainerTree([chain.audio, chain.allAudio]);
@@ -426,8 +431,14 @@ function addAudio(obj) {
     addCdsObject(obj, container);
 
     chain.genre.searchable = true;
-    container = addContainerTree([chain.audio, chain.allGenres, chain.genre]);
-    addCdsObject(obj, container);
+    if (obj.metaData[M_GENRE]) {
+        for (var oneGenre in obj.metaData[M_GENRE]) {
+            chain.genre.title = obj.metaData[M_GENRE][oneGenre];
+            chain.genre.metaData[M_GENRE] = [ oneGenre ];
+            container = addContainerTree([chain.audio, chain.allGenres, chain.genre]);
+            addCdsObject(obj, container);
+        }
+    }
 
     container = addContainerTree([chain.audio, chain.allYears, chain.year]);
     addCdsObject(obj, container);
@@ -546,7 +557,7 @@ function addAudioStructured(obj) {
     boxConfig.singleLetterBoxSize = 2 * boxConfig.divChar.length + 1;
 
     const chain = {
-        allArtists: { title: '-Artist-', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
+        allArtists: { title: '-Artist-', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER, metaData: [] },
         allGenres: { title: '-Genre-', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
         allAlbums: { title: '-Album-', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
         allTracks: { title: '-Track-', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
@@ -563,10 +574,12 @@ function addAudioStructured(obj) {
         date: { title: date, objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER }
     };
 
+    chain.allArtists.metaData[M_CONTENT_CLASS] = [ UPNP_CLASS_AUDIO_ITEM ];
     chain.album.metaData[M_ARTIST] = [ album_artist ];
     chain.album.metaData[M_ALBUMARTIST] = [ album_artist ];
     chain.album.metaData[M_GENRE] = [ genre ];
     chain.album.metaData[M_DATE] = obj.metaData[M_DATE];
+    chain.album.metaData[M_UPNP_DATE] = obj.metaData[M_UPNP_DATE];
     chain.album.metaData[M_ALBUM] = [ album ];
     chain.artist.metaData[M_ARTIST] = artist;
     chain.artist.metaData[M_ALBUMARTIST] = artist;
@@ -625,15 +638,27 @@ function addAudioStructured(obj) {
     obj.title = title + ' - ' + artist_full;
     chain.entryAll.title = '--all--';
     chain.entryAll.upnpclass = UPNP_CLASS_CONTAINER_MUSIC_GENRE;
-    container = addContainerTree([chain.allGenres, chain.genre, chain.entryAll]);
-    addCdsObject(obj, container);
+    if (obj.metaData[M_GENRE]) {
+        for (var oneGenre in obj.metaData[M_GENRE]) {
+            chain.genre.title = obj.metaData[M_GENRE][oneGenre];
+            chain.genre.metaData[M_GENRE] = [ oneGenre ];
+            container = addContainerTree([chain.allGenres, chain.genre, chain.entryAll]);
+            addCdsObject(obj, container);
+        }
+    }
 
     for (i = 0; i < artCnt; i++) {
         chain.abc.title = abcbox(artist[i], boxConfig.genre, boxConfig.divChar);
         isSingleCharBox = boxConfig.singleLetterBoxSize >= chain.abc.title.length;
         chain.album_artist.searchable = true;
-        container = addContainerTree(isSingleCharBox ? [chain.allGenres, chain.genre, chain.abc, chain.album_artist] : [chain.allGenres, chain.genre, chain.abc, chain.init, chain.album_artist]);
-        addCdsObject(obj, container);
+        if (obj.metaData[M_GENRE]) {
+            for (var oneGenre in obj.metaData[M_GENRE]) {
+                chain.genre.title = obj.metaData[M_GENRE][oneGenre];
+                chain.genre.metaData[M_GENRE] = [ oneGenre ];
+                container = addContainerTree(isSingleCharBox ? [chain.allGenres, chain.genre, chain.abc, chain.album_artist] : [chain.allGenres, chain.genre, chain.abc, chain.init, chain.album_artist]);
+                addCdsObject(obj, container);
+            }
+        }
     }
     chain.album_artist.searchable = false;
         
@@ -674,7 +699,7 @@ function addAudioStructured(obj) {
 function addVideo(obj) {
     const dir = getRootPath(object_script_path, obj.location);
     const chain = {
-        video: { title: 'Video', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
+        video: { title: 'Video', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER, metaData: [] },
         allVideo: { title: 'All Video', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
         allDirectories: { title: 'Directories', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
         allYears: { title: 'Year', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
@@ -684,6 +709,7 @@ function addVideo(obj) {
         month: { title: 'Unbekannt', objectType: OBJECT_TYPE_CONTAINER, searchable: true, upnpclass: UPNP_CLASS_CONTAINER, metaData: [], res: obj.res, aux: obj.aux, refID: obj.id },
         date: { title: 'Unbekannt', objectType: OBJECT_TYPE_CONTAINER, searchable: true, upnpclass: UPNP_CLASS_CONTAINER, metaData: [], res: obj.res, aux: obj.aux, refID: obj.id }
     };
+    chain.video.metaData[M_CONTENT_CLASS] = [ UPNP_CLASS_VIDEO_ITEM ];
     var container = addContainerTree([chain.video, chain.allVideo]);
     addCdsObject(obj, container);
     if (obj.metaData[M_CREATION_DATE] && obj.metaData[M_CREATION_DATE][0]) {
@@ -717,7 +743,7 @@ function addImage(obj) {
     const dir = getRootPath(object_script_path, obj.location);
 
     const chain = {
-        imageRoot: { title: 'Photos', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
+        imageRoot: { title: 'Photos', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER, metaData: [] },
         allImages: { title: 'All Photos', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
 
         allDirectories: { title: 'Directories', objectType: OBJECT_TYPE_CONTAINER, upnpclass: UPNP_CLASS_CONTAINER },
@@ -728,6 +754,7 @@ function addImage(obj) {
         month: { title: 'Unbekannt', objectType: OBJECT_TYPE_CONTAINER, searchable: true, upnpclass: UPNP_CLASS_CONTAINER, metaData: [], res: obj.res, aux: obj.aux, refID: obj.id },
         date: { title: 'Unbekannt', objectType: OBJECT_TYPE_CONTAINER, searchable: true, upnpclass: UPNP_CLASS_CONTAINER, metaData: [], res: obj.res, aux: obj.aux, refID: obj.id }
     };
+    chain.imageRoot.metaData[M_CONTENT_CLASS] = [ UPNP_CLASS_IMAGE_ITEM ];
     addCdsObject(obj, addContainerTree([chain.imageRoot, chain.allImages]));
     if (obj.metaData[M_DATE] && obj.metaData[M_DATE][0]) {
         var date = obj.metaData[M_DATE][0];
@@ -803,6 +830,7 @@ function addTrailer(obj) {
         if (reldate.length >= 7) {
             chain.date.title = reldate.slice(0, 7);
             chain.date.metaData[M_DATE] = [ reldate.slice(0, 7) ];
+            chain.date.metaData[M_UPNP_DATE] = [ reldate.slice(0, 7) ];
             addCdsObject(obj, addContainerTree([chain.trailerRoot , chain.appleTrailers, chain.relDate, chain.date]));
         }
     }
@@ -816,67 +844,79 @@ function addTrailer(obj) {
     if (postdate && postdate.length >= 7) {
         chain.date.title = postdate.slice(0, 7);
         chain.date.metaData[M_DATE] = [ postdate.slice(0, 7) ];
+        chain.date.metaData[M_UPNP_DATE] = [ postdate.slice(0, 7) ];
         addCdsObject(obj, addContainerTree([chain.trailerRoot , chain.appleTrailers, chain.postDate, chain.date]));
     }
 }
 // doc-add-trailer-end
 
 // doc-add-playlist-item-begin
-function addPlaylistItem(playlist_title, location, title, playlistChain, order, playlistOrder) {
+function addPlaylistItem(playlist_title, entry, playlistChain, playlistOrder) {
     // Determine if the item that we got is an URL or a local file.
 
-    if (location.match(/^.*:\/\//)) {
+    if (entry.location.match(/^.*:\/\//)) {
         var exturl = {};
 
         // Setting the mimetype is crucial and tricky... if you get it
         // wrong your renderer may show the item as unsupported and refuse
         // to play it. Unfortunately most playlist formats do not provide
         // any mimetype information.
-        exturl.mimetype = 'audio/mpeg';
+        exturl.mimetype = entry.mimetype ? entry.mimetype : 'audio/mpeg';
 
         // Make sure to correctly set the object type, then populate the
         // remaining fields.
         exturl.objectType = OBJECT_TYPE_ITEM_EXTERNAL_URL;
 
-        exturl.location = location;
-        exturl.title = (title ? title : location);
-        exturl.protocol = 'http-get';
-        exturl.upnpclass = UPNP_CLASS_ITEM_MUSIC_TRACK;
-        exturl.description = "Song from " + playlist_title;
+        exturl.location = entry.location;
+        exturl.title = (entry.title ? entry.title : entry.location);
+        exturl.protocol = entry.protocol ? entry.protocol : 'http-get';
+        exturl.upnpclass = exturl.mimetype.startsWith('video') ? UPNP_CLASS_VIDEO_ITEM : UPNP_CLASS_ITEM_MUSIC_TRACK;
+        exturl.description = entry.description ? entry.description : ("Entry from " + playlist_title);
 
         // This is a special field which ensures that your playlist files
         // will be displayed in the correct order inside a playlist
         // container. It is similar to the id3 track number that is used
         // to sort the media in album containers.
-        exturl.playlistOrder = (order ? order : playlistOrder++);
+        exturl.playlistOrder = (entry.order ? entry.order : playlistOrder);
 
         // Your item will be added to the container named by the playlist
         // that you are currently parsing.
         addCdsObject(exturl, playlistChain);
     } else {
-        if (location.substr(0,1) !== '/') {
-            location = playlistLocation + location;
+        if (entry.location.substr(0,1) !== '/') {
+            entry.location = playlistLocation + entry.location;
         }
-        var cds = getCdsObject(location);
+
+        var cds = getCdsObject(entry.location);
         if (!cds) {
-            print("Playlist " + playlist_title + " Skipping entry: " + location);
-            return playlistOrder;
+            print("Playlist " + playlist_title + " Skipping entry: " + entry.location);
+            return false;
         }
 
         var item = copyObject(cds);
 
-        item.playlistOrder = (order ? order : playlistOrder++);
+        item.playlistOrder = (entry.order ? entry.order : playlistOrder);
         item.title = item.metaData[M_TITLE] ? item.metaData[M_TITLE][0] : cds.title;
+        item.metaData[M_CONTENT_CLASS] = [ UPNP_CLASS_PLAYLIST_ITEM ];
+        item.description = entry.description ? entry.description : null;
+        // print("Playlist " + item.title + " Adding entry: " + item.playlistOrder + " " + item.location);
 
         addCdsObject(item, playlistChain);
     }
-    return playlistOrder;
+    return true;
 }
 // doc-add-playlist-item-end
 
 // doc-playlist-m3u-parse-begin
 function readM3uPlaylist(playlist_title, playlistChain, playlistDirChain) {
-    var title = null;
+    var entry = {
+        title: null,
+        location: null,
+        order: 0,
+        mimetype: null,
+        description: null,
+        protocol: null,
+    };
     var line = readln();
     var playlistOrder = 1;
 
@@ -887,32 +927,115 @@ function readM3uPlaylist(playlist_title, playlistChain, playlistDirChain) {
 
     // Here is the do - while loop which will read the playlist line by line.
     do {
-        var matches = line.match(/^#EXTINF:(-?\d+),\s?(\S.+)$/i);
+        var matches = line.match(/^#EXTINF:(-?\d+),\s*(\S.*)$/i);
+
         if (matches) {
             // duration = matches[1]; // currently unused
-            title = matches[2];
-        }
-        else if (!line.match(/^(#|\s*$)/)) {
+            entry.title = matches[2];
+            matches = line.match(/^#EXTINF:(-?\d+),\s*(\S.*),\s*(\S.*)$/i);
+            if (matches) {
+                entry.title = matches[2];
+                entry.mimetype = matches[3];
+            }
+        } else if (!line.match(/^(#|\s*$)/)) {
+            entry.location = line;
             // Call the helper function to add the item once you gathered the data:
-            playlistOrder = addPlaylistItem(playlist_title, line, title, playlistChain, 0, playlistOrder);
+            var state = addPlaylistItem(playlist_title, entry, playlistChain, playlistOrder);
 
             // Also add to "Directories"
-            if (playlistDirChain)
-                playlistOrder = addPlaylistItem(playlist_title, line, title, playlistDirChain, 0, playlistOrder);
+            if (playlistDirChain && state)
+                state = addPlaylistItem(playlist_title, entry, playlistDirChain, playlistOrder);
 
-            title = null;
+            entry.title = null;
+            entry.mimetype = null;
+            if (state)
+                playlistOrder++;
         }
-        
+
         line = readln();
     } while (line);
 }
 // doc-playlist-m3u-parse-end
 
+function readAsxPlaylist(playlist_title, playlistChain, playlistDirChain) {
+    var entry = {
+        title: null,
+        location: null,
+        order: 0,
+        mimetype: null,
+        description: null,
+        protocol: null,
+    };
+    var base = null;
+    var node = readXml(-2);
+    var playlistOrder = 1;
+    var level = 0;
+    
+    while (node || level > 0) {
+        if (!node && level > 0) {
+            node = readXml(-1); // read parent
+            node = readXml(0); // read next
+            level--;
+            if (entry.location) {
+                if (base)
+                    entry.location = base + '/' + entry.location;
+                var state = addPlaylistItem(playlist_title, entry, playlistChain, playlistOrder);
+                if (playlistDirChain && state)
+                    state = addPlaylistItem(playlist_title, entry, playlistDirChain, playlistOrder);
+                if (state)
+                    playlistOrder++;
+            }
+            entry.title = null;
+            entry.location = null;
+            entry.mimetype = null;
+            entry.protocol = null;
+            entry.description = null;
+        } else if (node.NAME === "asx") {
+            node = readXml(1); // read children
+            level++;
+        } else if (node.NAME === "entry") {
+            node = readXml(1); // read children
+            level++;
+        } else if (node.NAME === "ref" && node.href) {
+            entry.location = node.href;
+            node = readXml(0); // read nect
+        } else if (node.NAME === "base" && node.href && level < 2) {
+            base = node.href;
+            node = readXml(0); // read nect
+        } else if (node.NAME == "title") {
+            entry.title = node.VALUE;
+            node = readXml(0); // read nect
+        } else if (node.NAME == "abstract") {
+            entry.description = node.VALUE;
+            node = readXml(0); // read nect
+            // currently unused
+        } else if (node.NAME == "param" && node.name === "mimetype") {
+            entry.mimetype = node.value;
+            node = readXml(0); // read nect
+        } else if (node.NAME == "param" && node.name === "protocol") {
+            entry.protocol = node.value;
+            node = readXml(0); // read nect
+        } else {
+            node = readXml(0); // read nect
+        }
+    }
+
+    if (entry.location) {
+        var state = addPlaylistItem(playlist_title, entry, playlistChain, playlistOrder);
+        if (playlistDirChain && state)
+            addPlaylistItem(playlist_title, entry, playlistDirChain, playlistOrder);
+    }
+}
+
 function readPlsPlaylist(playlist_title, playlistChain, playlistDirChain) {
-    var title = null;
-    var file = null;
-    var lastId = -1;
-    var id = -1;
+    var entry = {
+        title: null,
+        location: null,
+        order: -1,
+        mimetype: null,
+        description: null,
+        protocol: null,
+    };
     var line = readln();
     var playlistOrder = 1;
     
@@ -926,48 +1049,74 @@ function readPlsPlaylist(playlist_title, playlistChain, playlistDirChain) {
             // var plsVersion =  RegExp.$1;
         } else if (line.match(/^File\s*(\d+)\s*=\s*(\S.+)$/i)) {
             const matches = line.match(/^File\s*(\d+)\s*=\s*(\S.+)$/i);
-            var thisFile = matches[2];
-            id = parseInt(matches[1], 10);
-            if (lastId === -1)
-                lastId = id;
-            if (lastId !== id) {
-                if (file) {
-                    playlistOrder = addPlaylistItem(playlist_title, file, title, playlistChain, lastId, playlistOrder);
-                    if (playlistDirChain)
-                        playlistOrder = addPlaylistItem(playlist_title, file, title, playlistDirChain, lastId, playlistOrder);
+            const thisFile = matches[2];
+            const id = parseInt(matches[1], 10);
+            if (entry.order === -1)
+                entry.order = id;
+            if (entry.order !== id) {
+                if (entry.location) {
+                    var state = addPlaylistItem(playlist_title, entry, playlistChain, playlistOrder);
+                    if (playlistDirChain && state)
+                        state = addPlaylistItem(playlist_title, entry, playlistDirChain, playlistOrder);
+                    if (state)
+                        playlistOrder++;
                 }
 
-                title = null;
-                lastId = id;
+                entry.title = null;
+                entry.mimetype = null;
+                entry.order = id;
             }
-            file = thisFile;
+            entry.location = thisFile;
         } else if (line.match(/^Title\s*(\d+)\s*=\s*(\S.+)$/i)) {
             const matches = line.match(/^Title\s*(\d+)\s*=\s*(\S.+)$/i);
-            var thisTitle = matches[2];
-            id = parseInt(matches[1], 10);
-            if (lastId === -1)
-                lastId = id;
-            if (lastId !== id) {
-                if (file) {
-                    playlistOrder = addPlaylistItem(playlist_title, file, title, playlistChain, lastId, playlistOrder);
-                    if (playlistDirChain)
-                        playlistOrder = addPlaylistItem(playlist_title, file, title, playlistDirChain, lastId, playlistOrder);
+            const thisTitle = matches[2];
+            const id = parseInt(matches[1], 10);
+            if (entry.order === -1)
+                entry.order = id;
+            if (entry.order !== id) {
+                if (entry.location) {
+                    var state = addPlaylistItem(playlist_title, entry, playlistChain, playlistOrder);
+                    if (playlistDirChain && state)
+                        state = addPlaylistItem(playlist_title, entry, playlistDirChain, playlistOrder);
+                    if (state)
+                        playlistOrder++;
                 }
 
-                file = null;
-                lastId = id;
+                entry.location = null;
+                entry.mimetype = null;
+                entry.order = id;
             }
-            title = thisTitle;
+            entry.title = thisTitle;
+        } else if (line.match(/^MimeType\s*(\d+)\s*=\s*(\S.+)$/i)) {
+            const matches = line.match(/^MimeType\s*(\d+)\s*=\s*(\S.+)$/i);
+            const thisMime = matches[2];
+            const id = parseInt(matches[1], 10);
+            if (entry.order === -1)
+                entry.order = id;
+            if (entry.order !== id) {
+                if (entry.location) {
+                    var state = addPlaylistItem(playlist_title, entry, playlistChain, playlistOrder);
+                    if (playlistDirChain && state)
+                        state = addPlaylistItem(playlist_title, entry, playlistDirChain, playlistOrder);
+                    if (state)
+                        playlistOrder++;
+                }
+
+                entry.location = null;
+                entry.title = null;
+                entry.order = id;
+            }
+            entry.mimetype = thisMime;
         } else if (line.match(/^Length\s*(\d+)\s*=\s*(\S.+)$/i)) {
             // currently unused
         }
 
         line = readln();
-    } while (line);
+    } while (line !== undefined);
 
-    if (file) {
-        playlistOrder = addPlaylistItem(playlist_title, file, title, playlistChain, lastId, playlistOrder);
-        if (playlistDirChain)
-            playlistOrder = addPlaylistItem(playlist_title, file, title, playlistDirChain, lastId, playlistOrder);
+    if (entry.location) {
+        var state = addPlaylistItem(playlist_title, entry, playlistChain, playlistOrder);
+        if (playlistDirChain && state)
+            addPlaylistItem(playlist_title, entry, playlistDirChain, playlistOrder);
     }
 }
